
lib_public {

    // ecol.l                      surf.l-based edge collapse library
    // by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
    //
    // 01-06-00: wrote and debugged non-prioritized collapse functions
    // 01-07-00: split from mesh.l, added priority-queue collapsing
    // 01-08-00: debugged and tested priority collapses
    // 01-09-00: added ecol_do_remove() for post-collapse base-mesh cleanup
    // 02-19-00: moved ecol.l into Surf/surf.l archive


    // KNOWN FAILURE CONDITIONS
    // ------------------------
    //
    // FIXED ==> (1) will seg fault if there is a mesh boundary
    // (2) edge with two verts total valence >256 will cause buffer overrun
    // FIXED ==> (3) perhaps rare exit(1) on valence 3 stack overflow;
    //     this might not be possible--it never came up in extensive testing.


    lib_include "Geo/geo.l";
    lib_include "../surf.l";


    #define ECQUEUE 0x100000


    lib_struct {
        edge e;                // edge (represented by end e with e<e->e)
        int iq;                // queue priority index
        ecol ecq0,ecq1;        // links in queue's bucket iq
        float eacc,emax;       // accumulated error, max additional error
        ecol ech0,ech1;        // links in hash bucket
    } ecol;


    void ecol_do_collapse(surf s,int nv_target);
    void ecol_collapse_label(surf s);
    void ecol_do_remove(surf s,double flatness);
    void ecol_do_triangulate(surf s);

};


typedef int (*ecolpriority)(ecol ec);


void ecol_bgn_optimize(surf s,ecolpriority priority);
void ecol_end_optimize();
void ecol_update(edge e,double err);
void ecol_queue(ecol ec,int iq);
void ecol_destroy(edge e);
int ecol_hash(edge *re);
ecol ecol_find(edge *re,int *rh);
ecol ecol_get(edge e);

int ecol_collapse_priority(ecol ec);
int ecol_collapse(ecol ec);
void ecol_collapse_point(edge e,double *rxc,double *ryc,double *rzc);
int ecol_endissharp(edge ex);
int ecol_collapse_okay(surf s,edge e,edge *etab,int *rne);

int ecol_remove_priority(ecol ec);
int ecol_remove(ecol ec);


#define ABS(X) (X<0.0?-X:X)

#define NORM(X,Y,Z) {                                                         \
    r=X*X+Y*Y+Z*Z;                                                            \
    if (r>0.0) { r=sqrt(r); X/=r; Y/=r; Z/=r; }                               \
    else { r=1e-8; X=1.0; Y=0.0; Z=0.0; }                                     \
}

#define CROSS(X,Y,Z,X0,Y0,Z0,X1,Y1,Z1) \
    { X=Y0*Z1-Z0*Y1; Y=Z0*X1-X0*Z1; Z=X0*Y1-Y0*X1; }


surf ec_s;                                  // surf being collapsed
ecolpriority ec_priority;                   // current priority function
mempool ec_mp;                              // pool for ecol allocation
ecol ec_queue0[ECQUEUE],ec_queue1[ECQUEUE]; // queue bucket ends
int ec_iqmax;                               // max-priority occupied bucket
ecol *ec_hashtab0,*ec_hashtab1;             // hash bucket ends
int ec_hmax;                                // hash table size (power of two)
int ec_nv;                                  // current vertex count
double ec_r0,ec_r1;                         // queue log() limits
double ec_flatness;                         // tolerance for _remove()'s


void ecol_do_collapse(surf s,int nv_target)
{
    int i,nv,ne,nf,ei;
    edge e;
    ecol ec;

    printf("ecol_do_collapse {\n"); fflush(stdout);

    // do generic optimization setup
    ecol_bgn_optimize(s,ecol_collapse_priority);

    // make requested number of collapses
    printf("ecol_do_collapse 1: ec_iqmax=%d\n",ec_iqmax); fflush(stdout);
    if (ec=ec_queue0[ec_iqmax]) {
        printf("  ec_nv=%d nv_target=%d ec_queue0[%d]=%8.8x err=%g+%g\n",
            ec_nv,nv_target,ec_iqmax,ec,ec->eacc,ec->emax); fflush(stdout);
    }

    {
        double dt0;

        dt0=timer_systemdsec();

        while (ec_nv>nv_target && (ec=ec_queue0[ec_iqmax])) {
            if (ecol_collapse(ec)) {
                // fix incorrectly-queued not-okay edge
                // (tends to occur only near full collapse)
                ecol_update(ec->e,0.0);
                printf("  +++ fixed queued not-okay edge\n");
            }
            if (ec_nv<200) {
                printf(
                "  >> ec_nv=%d nv_target=%d ec_queue0[%d]=%8.8x err=%g+%g\n",
                    ec_nv,nv_target,ec_iqmax,ec,ec->eacc,ec->emax);
                fflush(stdout);
            }
        }
        dt0=timer_systemdsec()-dt0;
        printf("   collapse time: %g sec\n",dt0); fflush(stdout);
    }

    // do generic optimization teardown
    ecol_end_optimize();
    printf("ecol_do_collapse }\n"); fflush(stdout);
}


void ecol_collapse_label(surf s)
{
    int nv,ne,nf,ei,nex;
    edge e,*edges,etab[256];

    surf_size(s,&nv,&ne,&nf);
    edges=(edge *)mem_get(ne*sizeof(edge),"ecol");
    ei=0;
    BGNEDGES
        edges[ei++]=(surf_etoi(e)<surf_etoi(e->e)?e:(edge)0);
    ENDEDGES

    for (ei=0;ei<ne;ei++) {
        if (e=edges[ei]) {
            if (ecol_collapse_okay(s,e,etab,&nex)) edge_setstate(e,1);
            else edge_setstate(e,0);
        }
    }
    mem_put((char *)edges,"ecol");
}


void ecol_do_remove(surf s,double flatness)
{
    int i;
    ecol ec;

    printf("ecol_do_remove {\n"); fflush(stdout);

    ec_flatness=flatness;
    ecol_bgn_optimize(s,ecol_remove_priority);

    printf("ecol_do_remove 1\n"); fflush(stdout);
    i=0;
    while (ec=ec_queue0[ec_iqmax]) i+=ecol_remove(ec);

    printf("ecol_do_remove 2\n"); fflush(stdout);
    ecol_end_optimize(s,ecol_remove_priority);

    printf("ecol_do_remove }\n"); fflush(stdout);
}

void ecol_do_triangulate(surf s)
{
    int nv,ne,nf,fi,i,j,cv;
    face f,*ftab;
    edge e,e0,e1,ex;
    vert vtab[256];

    surf_size(s,&nv,&ne,&nf);
    ftab=(face *)mem_get(nf*sizeof(face),"ecol");
    fi=0;
    BGNFACES
        ftab[fi]=f;

        // check if any verts are duplicated
        cv=0;
        for (e=e0=f->e0;e;e=e1) {
            e1=e->e->ve0; if (e1==e0) e1=(edge)0;
            vtab[cv++]=e->v;
        }
        for (i=0;i+1<cv;i++) {
            for (j=i+1;j<cv;j++) {
                if (vtab[i]==vtab[j]) {
                    fprintf(stderr,"*** double vert in face\n"); exit(1);
                }
            }
        }
        fi++;
    ENDFACES
    for (fi=0;fi<nf;fi++) {
        f=ftab[fi];
        if (f->e0->e->ve0->e->ve0==f->e0) {
            fprintf(stderr,"*** triangulate: two-edged face\n"); exit(1);
        }
        if (f->e0->e->ve0->e->ve0->e->ve0!=f->e0) {
            e=f->e0;
            face_destroy(f);
            for (e0=e,e=e->e->ve0->e->ve0;e;e=e1) {
                e1=e->e->ve0; if (e1->e->v==e0->v) e1=(edge)0;
                ex=edge_create(e0->v,e->v);
                vert_insertedge(ex,e->ve1->e->ve1->e,(edge)0);
                vert_insertedge(ex->e,e,(edge)0);
                f=face_create(ex->e);
                if (!e1) f=face_create(ex);
            }
        }
    }

    mem_put((char *)ftab,"ecol");
}


// ========== internal routines ==========


void ecol_bgn_optimize(surf s,ecolpriority priority)
{
    int i,nv,ne,nf;
    edge e;

    // set up pool, queue and hashtab, add okay edges
    ec_s=s;
    ec_priority=priority;
    surf_size(s,&nv,&ne,&nf);
    ec_nv=nv;
    ec_mp=mempool_create(sizeof(ecol_struct),(char *)0);
    for (i=0;i<ECQUEUE;i++) ec_queue0[i]=ec_queue1[i]=(ecol)0;
    ec_iqmax=0;
    ec_hmax=ne; ec_hmax|=ec_hmax>>1; ec_hmax|=ec_hmax>>2; ec_hmax|=ec_hmax>>4;
    ec_hmax|=ec_hmax>>8; ec_hmax|=ec_hmax>>16; ec_hmax=(ec_hmax+1)>>2;
    if (ec_hmax<4096) ec_hmax=4096;
    ec_hashtab0=(ecol *)mem_get(ec_hmax*sizeof(ecol),"ecol");
    ec_hashtab1=(ecol *)mem_get(ec_hmax*sizeof(ecol),"ecol");
    for (i=0;i<ec_hmax;i++) ec_hashtab0[i]=ec_hashtab1[i]=(ecol)0;
    ec_r0=log(1.0/1e10); ec_r1=log(1e10);
    i=0;
    BGNEDGES
        if (i<surf_etoi(e->e)) ecol_update(e,0.0);
        i++;
    ENDEDGES
}


void ecol_end_optimize()
{
    // de-allocate pool, hashtabs
    mempool_destroy(ec_mp);
    mem_put((char *)ec_hashtab0,"ecol");
    mem_put((char *)ec_hashtab1,"ecol");
}


void ecol_update(edge e,double err)
{
    int iq;
    ecol ec;

    ec=ecol_get(e); if (err>ec->eacc) ec->eacc=err;
    iq=(*ec_priority)(ec);
    ecol_queue(ec,iq);
}


void ecol_queue(ecol ec,int iq)
{
    // return if already in correct queue location
    if (ec->iq==iq) return;

    // dequeue first if in queue
    if (ec->iq>=0) {
        if (ec->ecq0) ec->ecq0->ecq1=ec->ecq1;
        else ec_queue0[ec->iq]=ec->ecq1;
        if (ec->ecq1) ec->ecq1->ecq0=ec->ecq0;
        else ec_queue1[ec->iq]=ec->ecq0;
        if (ec->iq==ec_iqmax)
            { while (ec_iqmax>0 && !ec_queue0[ec_iqmax]) ec_iqmax--; }
        ec->iq= -1;
    }

    // set queue index
    ec->iq=iq;

    // enter on queue if requested
    if (iq>=0) {
        ec->ecq0=ec_queue1[iq]; ec->ecq1=(ecol)0;
        if (ec->ecq0) ec->ecq0->ecq1=ec; else ec_queue0[iq]=ec;
        if (ec->ecq1) ec->ecq1->ecq0=ec; else ec_queue1[iq]=ec;
        if (iq>ec_iqmax) ec_iqmax=iq;
    }
}


void ecol_destroy(edge e)
{
    int h;
    ecol ec;

    if (!(ec=ecol_find(&e,&h))) return;
    ecol_queue(ec,-1);
    if (ec->ech0) ec->ech0->ech1=ec->ech1; else ec_hashtab0[h]=ec->ech1;
    if (ec->ech1) ec->ech1->ech0=ec->ech0; else ec_hashtab1[h]=ec->ech0;
    mempool_put((char *)ec);
}


int ecol_hash(edge *re)
{
    int i,j,h;
    edge e;

    e= *re;
    i=e-(edge)0; j=e->e-(edge)0;
    if (j<i) { i=j; *re=e->e; }
    h=((i+0xcafefeed)>>16)^((i+0xdeadbeef)>>8)^(i+0xbabeface);
    h&=ec_hmax-1;
    return h;
}


ecol ecol_find(edge *re,int *rh)
{
    int h;
    edge e;
    ecol ec;

    h=ecol_hash(re);
    e= *re;
    for (ec=ec_hashtab0[h];ec && ec->e!=e;ec=ec->ech1) ;
    *rh=h;
    return ec;
}


ecol ecol_get(edge e)
{
    int h;
    ecol ec;

    if (!(ec=ecol_find(&e,&h))) {
        // make new ec since none found
        ec=(ecol)mempool_get(ec_mp);
        ec->e=e;
        ec->iq= -1; // unqueued
        ec->ecq0=ec->ecq1=(ecol)0;
        ec->eacc=0.0; ec->emax=0.0;
        ec->ech0=ec_hashtab1[h]; ec->ech1=(ecol)0;
        if (ec->ech0) ec->ech0->ech1=ec; else ec_hashtab0[h]=ec;
        if (ec->ech1) ec->ech1->ech0=ec; else ec_hashtab1[h]=ec;
    }
    return ec;
}


// ---------- _collapse support ----------


int ecol_collapse_priority(ecol ec)
{
    int i,ne,iq;
    double r,dx,dy,dz,x0,y0,z0,x1,y1,z1,nx,ny,nz,emax,xc,yc,zc;
    double nxtab[16],nytab[16],nztab[16],ax,ay,az,rmax,rmin;
    vert v0,v1;
    edge e,ex,e0,e1,etab[256];
    surf s;

    e=ec->e; s=surf_findsurf((char *)e);

    // disallow collapse of edge with two sharp ends
    if (ecol_endissharp(e) && ecol_endissharp(e->e)) return -1;

    // if not okay return iq= -1 (dequeued)
    if (!ecol_collapse_okay(s,e,etab,&ne)) return -1;

    // update ec->emax
    v0=e->v; v1=e->e->v;
    dx=v1->x-v0->x; dy=v1->y-v0->y; dz=v1->z-v0->z;

    //xc=(v0->x+v1->x)/2.0; yc=(v0->y+v1->y)/2.0; zc=(v0->z+v1->z)/2.0;
    ecol_collapse_point(e,&xc,&yc,&zc);

    ne=0;
    ex=e;
    for (e0=ex->ve0,ex=ex->ve1;ex;ex=e1) {
        e1=ex->ve1; if (e1==e0) e1=(edge)0;
        etab[ne++]=ex->e->ve0;
    }
    ex=e->e;
    for (e0=ex->ve0,ex=ex->ve1;ex;ex=e1) {
        e1=ex->ve1; if (e1==e0) e1=(edge)0;
        etab[ne++]=ex->e->ve0;
    }
    emax=0.0; ax=ay=az=0.0; rmax=1e-8;
    for (i=0;i<ne;i++) {
        if (etab[i]->f) {
            v0=etab[i]->v; v1=etab[i]->e->v;
            x0=v0->x-xc; y0=v0->y-yc; z0=v0->z-zc;
            x1=v1->x-xc; y1=v1->y-yc; z1=v1->z-zc;
            CROSS(nx,ny,nz,x0,y0,z0,x1,y1,z1)
            nxtab[i]=nx; nytab[i]=ny; nztab[i]=nz;
            NORM(nx,ny,nz)
            ax+=nx; ay+=ny; az+=nz; if (r>rmax) rmax=r;
            r=nx*dx+ny*dy+nz*dz;
            if (r<0.0) r= -r;
            if (r>emax) emax=r;
        }
    }
    if (emax<1e-8) emax=1e-8; // this allows distinction in planar areas
    NORM(ax,ay,az) rmin=1e20;
    for (i=0;i<ne;i++) {
        if (etab[i]->f) {
            r=nxtab[i]*ax+nytab[i]*ay+nztab[i]*az;
            if (r<rmin) rmin=r;
        }
    }
    //rmin=2.0-rmin/rmax; rmin*=rmin;
    rmin=2.0-rmin/rmax; rmin*=rmin; rmin*=rmin; rmin*=rmin; rmin*=rmin;
    ec->emax=rmin*emax/2.0;

    // compute and return iq
    r=ec->eacc+ec->emax;
    r=(r>1e-20?1.0/r:1e20);
    r=log(r);
    r=(r-ec_r0)/(ec_r1-ec_r0);
    iq=(int)floor(r*(double)ECQUEUE);
    if (iq<0) iq=0; if (iq>=ECQUEUE) iq=ECQUEUE-1;
    return iq;
}


int ecol_collapse(ecol ec)
{
    int i,j,ne,esharp,vsharp;
    edge e,ex,ey,ez,ew,e0,e1,etab[256],ea,eb,eup[2];
    vert v,v0,v1;
    surf s;
    double err,x,y,z;

    // get info on edge
    e=ec->e; s=surf_findsurf((char *)e);
    err=ec->eacc+ec->emax;
    esharp=edge_issharp(e);
    vsharp=(vert_issharp(e->v) || vert_issharp(e->e->v));

    // quit if not okay to collapse
    if (!ecol_collapse_okay(s,e,etab,&ne)) return 1;

    // create new center vertex
    v0=e->v; v1=e->e->v;
    ecol_collapse_point(e,&x,&y,&z);
    if (vsharp) surf_sharp();
    v=vert_create(s,x,y,z);
    ec_nv++;

    // disengage edge and it's verts and destroy them
    if (e->f && e->f->e0==e) e->f->e0=e->e->ve0;
    if (e->e->f && e->e->f->e0==e->e) e->e->f->e0=e->ve0;
    for (i=0;i<ne;i++) {
        ex=etab[i];
        if (ex!=e && ex!=e->e) {
            ex->v=v;
            if (!v->e0) v->e0=ex->ve0=ex->ve1=ex;
            else{
                e1=v->e0; e0=e1->ve0;
                ex->ve0=e0; ex->ve1=e1; e0->ve1=e1->ve0=ex;
            }
        }
    }
    e->ve0=e->ve1=(edge)0;
    e->e->ve0=e->e->ve1=(edge)0;
    e->f=(face)0; e->e->f=(face)0;
    v0->e0=(edge)0; v1->e0=(edge)0;
    ecol_destroy(e); edge_destroy(e);
    vert_destroy(v0); vert_destroy(v1);
    ec_nv-=2;

    // record new edge ring
    ne=0;
    for (e=e0=v->e0;e;e=e1) {
        e1=e->ve1; if (e1==e0) e1=(edge)0;
        etab[ne++]=e;
    }

    // eliminate two-edge faces
    eup[0]=eup[1]=(edge)0;
    for (i=0;i<ne;i++) {
        e=etab[i];
        ex=e->ve1;
        if (e->e->ve0->e==ex) {

            e0=e; ec=ecol_find(&e0,&j);
            if (ec->eacc+ec->emax>err) err=ec->eacc+ec->emax;
            e0=ex; ec=ecol_find(&e0,&j);
            if (ec->eacc+ec->emax>err) err=ec->eacc+ec->emax;
            ecol_destroy(e); ecol_destroy(ex);
            if (e->f) face_destroy(e->f);

            ey=e->e; ea=ex->ve0; eb=ey->ve0;
            if (edge_issharp(ex) || edge_issharp(ey)) surf_sharp();
            ez=edge_create(ex->v,ey->v); ew=ez->e;

            j=i+1; if (j>=ne) j=0;
            if (etab[j]!=ex) { printf("oops! etab[j]!=ex\n"); exit(1); }
            etab[i]=(edge)0; etab[j]=ez;

            ez->ve0=ea->ve0; ez->ve1=ex->ve1; ez->f=ex->f;
            ew->ve0=eb->ve0; ew->ve1=ey->ve1; ew->f=ey->f;
            ez->ve0->ve1=ez; ez->ve1->ve0=ez;
            ew->ve0->ve1=ew; ew->ve1->ve0=ew;
            ez->v->e0=ez; if (ez->f) ez->f->e0=ez;
            ew->v->e0=ew; if (ew->f) ew->f->e0=ew;

            ea->ve0=ea->ve1=eb->ve0=eb->ve1=(edge)0;
            ex->ve0=ex->ve1=ey->ve0=ey->ve1=(edge)0;
            ea->f=eb->f=ex->f=ey->f=(face)0;
            edge_destroy(ex); edge_destroy(ey);
            ex=ez; ey=ew;

            if (eup[0]) eup[1]=ex; else eup[0]=ex;
        }
    }

    // update errors/priorities in neighborhood
    for (i=0;i<ne;i++) {
        if (e=etab[i]) {
            e=e->e;
            ecol_update(e,err);
            ecol_update(e->ve0,0.0);
        }
    }
    for (i=0;i<2;i++) {
        if (e=eup[i]) {
            e=e->e;
            ecol_update(e->ve1->e->ve0,0.0);
            for (e0=e->ve0,e=e->ve1->ve1;e;e=e1) {
                e1=e->ve1; if (e1==e0) e1=(edge)0;
                ecol_update(e,0.0);
                ecol_update(e->e->ve0,0.0);
            }
        }
    }

    return 0;
}


void ecol_collapse_point(edge e,double *rxc,double *ryc,double *rzc)
{
    vert v0,v1;

    v0=e->v; v1=e->e->v;
    if (ecol_endissharp(e)) { *rxc=v0->x; *ryc=v0->y; *rzc=v0->z; return; }
    if (ecol_endissharp(e->e)) { *rxc=v1->x; *ryc=v1->y; *rzc=v1->z; return; }
    *rxc=(v0->x+v1->x)/2.0; *ryc=(v0->y+v1->y)/2.0; *rzc=(v0->z+v1->z)/2.0;
}


int ecol_endissharp(edge ex)
{
    int i;
    edge e,e0,e1;

    if (vert_issharp(ex->v)) return 1;
    i=0;
    for (e0=ex,e=ex->ve1;e;e=e1) {
        e1=e->ve1; if (e1==e0) e1=(edge)0;
        if (edge_issharp(e)) i++;
    }
    return (edge_issharp(ex)?0:i);
}


int ecol_collapse_okay(surf s,edge e,edge *etab,int *rne)
{
    int i,j,ne,ne0,ne1,dupcnt,tri0,tri1,cne;
    edge ex,e0,e1;

    ne=0; ne0=ne1=0;
    for (ex=e0=e;ex;ex=e1) {
        e1=ex->ve1; if (e1==e0) e1=(edge)0;
        etab[ne++]=ex;
    }
    for (ex=e0=e->e;ex;ex=e1) {
        e1=ex->ve1; if (e1==e0) e1=(edge)0;
        etab[ne++]=ex;
    }
    dupcnt=0;
    for (i=0;i+1<ne;i++) {
        for (j=i+1;j<ne;j++) {
            if (etab[i]->e->v==etab[j]->e->v) dupcnt++;
        }
    }
    tri0=tri1=0;
    if (e->e->ve0->e->ve0->e->ve0==e) {
        tri0=1;
        for (e0=ex=e->ve0->e;ex;ex=e1) {
            e1=ex->ve1; if (e1==e0) e1=(edge)0;
            ne0++;
        }
    }
    if (e->ve0->e->ve0->e->ve0->e==e) {
        tri1=1;
        for (e0=ex=e->ve1->e;ex;ex=e1) {
            e1=ex->ve1; if (e1==e0) e1=(edge)0;
            ne1++;
        }
    }
    *rne=ne;
    cne=ne-(tri0+tri1+2);
    //return (cne<9 && dupcnt==tri0+tri1 && cne>3);
    return (cne<9 && dupcnt==tri0+tri1 &&
            cne > ( (!e->f || !e->e->f) ? 2 : 3 ) );
}


// ---------- _remove support ----------


int ecol_remove_priority(ecol ec)
{
    int iq,nf,nf0,nf1,nv0,nv1;
    double r,x,y,z,x0,y0,z0,x1,y1,z1;
    double d,ax,ay,az,bx,by,bz;
    vert v;
    edge e,e0,e1,ex,ey;

    ex=ec->e; ey=ex->e;

    if (edge_issharp(ex) || !ex->f || !ey->f) return -1;

    // face valences and centers
    nf0=0; x=y=z=0.0;
    for (e0=e=ex;e;e=e1) {
        e1=e->e->ve0; if (e1==e0) e1=(edge)0;
        nf0++; v=e->v; x+=v->x; y+=v->y; z+=v->z;
    }
    r=1.0/(double)nf0; x0=r*x; y0=r*y; z0=r*z;
    nf1=0; x=y=z=0.0;
    for (e0=e=ey;e;e=e1) {
        e1=e->e->ve0; if (e1==e0) e1=(edge)0;
        nf1++; v=e->v; x+=v->x; y+=v->y; z+=v->z;
    }
    r=1.0/(double)nf1; x1=r*x; y1=r*y; z1=r*z;
    nf=nf0+nf1-2; if (nf>5) return -1;

    // vertex valences
    nv0=0;
    for (e0=e=ex;e;e=e1) {
        e1=e->ve1; if (e1==e0) e1=(edge)0;
        nv0++;
    }
    if (nv0-1<3) return -1;
    nv1=0;
    for (e0=e=ey;e;e=e1) {
        e1=e->ve1; if (e1==e0) e1=(edge)0;
        nv1++;
    }
    if (nv1-1<3) return -1;

    // face normals
    v=ex->v; ax=v->x-x0; ay=v->y-y0; az=v->z-z0;
    v=ey->v; bx=v->x-x0; by=v->y-y0; bz=v->z-z0;
    CROSS(x0,y0,z0,ax,ay,az,bx,by,bz) NORM(x0,y0,z0)
    v=ey->v; ax=v->x-x1; ay=v->y-y1; az=v->z-z1;
    v=ex->v; bx=v->x-x1; by=v->y-y1; bz=v->z-z1;
    CROSS(x1,y1,z1,ax,ay,az,bx,by,bz) NORM(x1,y1,z1)

    // base priority
    d=x0*x1+y0*y1+z0*z1;
    if (d<ec_flatness) return -1;
    d=1.0/(1.000001-d); d*=d;

    // multiply by modifiers
    d*=(double)nv0-3.5;
    d*=(double)nv1-3.5;
    if (nf>4) d/=(double)nf-3.5;

    // convert into queue-index return value
    if (d<1e-8) d=1e-8; 
    d=log(d);
    d=(d-ec_r0)/(ec_r1-ec_r0);
    iq=(int)floor(d*(double)ECQUEUE);
    if (iq<0) iq=0; if (iq>=ECQUEUE) iq=ECQUEUE-1;
    return iq;
}

int ecol_remove(ecol ec)
{
    edge e,ex,ey,e0,e1;
    face f;

    e=ec->e; ex=e->ve0; ey=e->e->ve0;
    ecol_destroy(e); edge_destroy(e);
    f=face_create(ex);
    for (e0=e=f->e0;e;e=e1) {
        e1=e->e->ve0; if (e1==e0) e1=(edge)0;
        ecol_update(e,0.0);
    }
    for (e0=ex,e=ex->ve1->ve1;e;e=e1) {
        e1=e->ve1; if (e1==e0) e1=(edge)0;
        ecol_update(e,0.0);
    }
    for (e0=ey,e=ey->ve1->ve1;e;e=e1) {
        e1=e->ve1; if (e1==e0) e1=(edge)0;
        ecol_update(e,0.0);
    }
    return 0;
}


