
// bof2mes.l      computes isosurface mesh from brick-of-float (.bof) files
// by Mark Duchaineau (free but copyrighted, see LibGen/COPYING)
//
// 01-18-97: wrote
// 12-04-99: added shrink arg, clean up for base-mesh generation, shrink-wrap
// 01-03-00: changed to be bof2mes.l instead of bob2mes.l for high quality
// 01-05-00: made iso extract temps sparse
// 01-09-00: updated to use ecol.l base-mesh generation
// 01-10-00: allowed mesh with boundaries (erase brick boundaries)
// 10-02-00: reduce/refine args
// 10-03-00: orig surf edge aging for viswin plotting
// 10-04-00: moved shrink-wrap stuff elsewhere

lib_include "Sys/sys.l";
lib_include "Surf/surf.l";


void getsample(int xi,int yi,int zi,
    double *rx,double *ry,double *rz,double *rd);
vert vhash_find(int i);
void addedgesub(surf s,int rev,vert v,int i1,int i2,int i3);
void bofshrink(int shrink,
    float **rbof,int *rxsize,int *rysize,int *rzsize);
void bofshrinksub(float **rbof,int *rxsize,int *rysize,int *rzsize);

float *brick,*brick0,*atab;
int xsize,ysize,zsize,xsize0,ysize0,zsize0,shrink;
double xc,yc,zc,qq,isolevel,qq0,qq1;
char *boffile;
int dont_put_brick;


#define ABS(X) (X<0.0?-X:X)

#define NORM(X,Y,Z) {                                                         \
    r=X*X+Y*Y+Z*Z;                                                            \
    if (r>0.0) { r=sqrt(r); X/=r; Y/=r; Z/=r; }                               \
    else { r=1e-8; X=1.0; Y=0.0; Z=0.0; }                                     \
}

#define CROSS(X,Y,Z,X0,Y0,Z0,X1,Y1,Z1) \
    { X=Y0*Z1-Z0*Y1; Y=Z0*X1-X0*Z1; Z=X0*Y1-Y0*X1; }



#define WVERTI(XI,YI,ZI,D) \
    (((((ZI)+1)*(ysize+2)+((YI)+1))*(xsize+2)+((XI)+1))*3+(D))

lib_struct {
    int i;       // unique vertex index within volume
    vhash vh1;   // next vertex in vhashtab list
    vert v;      // the vert
} vhash;

int h,hmax;
vhash *vhashtab;
vert v_dummy;

lib_main {
    int xi,yi,zi,i,n,dxi,dyi,dzi,rev,shrink,reduce,refine;
    int nv,ne,nf;
    double x0,y0,z0,d0,x1,y1,z1,d1,x,y,z,a,b,dx,dy,dz,r,rmax;
    surf s;
    vert v,v1,*xv,*yv,*zv;
    edge e,ex;
    face f;
    char *surffile;

    // get command-line arguments
    if (argc!=8) {
        printf("usage: bof2mes boffile nx ny nz isolevel out.mesb shrink\n");
        printf("    shrink=number of times to have 2x2x2 average to 1\n");
        exit(0);
    }
    boffile=argv[1];
    xsize=ysize=zsize= -1;
    isolevel=0.0;
    sscanf(argv[2],"%d",&xsize);
    sscanf(argv[3],"%d",&ysize);
    sscanf(argv[4],"%d",&zsize);
    sscanf(argv[5],"%lf",&isolevel);
    surffile=argv[6];
    sscanf(argv[7],"%d",&shrink);

    // read brick, shrink as requested
    printf("read brick, shrink as requested [argc=%d shrink=%d]\n",
        argc,shrink); fflush(stdout);
    brick=(float *)fio_read(boffile);
    if (!brick) { fprintf(stderr,"can't read %s\n",boffile); exit(1); }
    n=mem_size((char *)brick);
    brick0=brick; dont_put_brick=1;
    xsize0=xsize; ysize0=ysize; zsize0=zsize;
    if (shrink>=0) bofshrink(shrink,&brick,&xsize,&ysize,&zsize);

    // get translate/scale coefficients
    xc=(double)(xsize-1)/2.0; qq=xc;
    yc=(double)(ysize-1)/2.0; if (yc>qq) qq=yc;
    zc=(double)(zsize-1)/2.0; if (zc>qq) qq=zc;
    qq=1.0/qq; if (shrink>=0) { qq1=(double)(1<<shrink); qq0=(qq1-1.0)/2.0; }

    // extract isosurface as surf s
    {
        mempool vhmp;
        vhash vh;

        // create empty surface
        s=surf_create();
        v_dummy=vert_create(s,0.0,0.0,0.0);

        // number of vertex indices
        n=(xsize+2)*(ysize+2)*(zsize+2)*3;

        // good vhashtab size
        hmax=n; hmax|=hmax>>1; hmax|=hmax>>2; hmax|=hmax>>4;
        hmax|=hmax>>8; hmax|=hmax>>16; hmax++; hmax>>=5;
        if (hmax<4096) hmax=4096;

        // allocate temporary tables
        vhashtab=(vhash *)mem_get(hmax*sizeof(vhash),"bof2mes");
        for (h=0;h<hmax;h++) vhashtab[h]=(vhash)0;
        vhmp=mempool_create(sizeof(vhash_struct),(char *)0);

        // compute verts
        //   for every zone edge, if +/- or -/+ then create vert
        printf("compute verts\n"); fflush(stdout);
        #define COMPVERTS(X0,Y0,Z0,DX,DY,DZ,D) {                              \
            for (xi=0;xi+DX<xsize;xi++) {                                     \
                for (yi=0;yi+DY<ysize;yi++) {                                 \
                    for (zi=0;zi+DZ<zsize;zi++) {                             \
                        getsample(xi,yi,zi,&x0,&y0,&z0,&d0);                  \
                        getsample(xi+DX,yi+DY,zi+DZ,&x1,&y1,&z1,&d1);         \
                        if ((d0<isolevel && d1>=isolevel) ||                  \
                            (d1<isolevel && d0>=isolevel)) {                  \
                            b=(isolevel-d0)/(d1-d0); a=1.0-b;                 \
                            x=a*x0+b*x1; y=a*y0+b*y1; z=a*z0+b*z1;            \
                            v=vert_create(s,x,y,z);                           \
                            i=WVERTI(xi,yi,zi,(D)); h=i&(hmax-1);             \
                            vh=(vhash)mempool_get(vhmp);                      \
                            vh->i=i; vh->v=v;                                 \
                            vh->vh1=vhashtab[h]; vhashtab[h]=vh;              \
                        }                                                     \
                    }                                                         \
                }                                                             \
            }                                                                 \
        }
        COMPVERTS(-1, 0, 0,1,0,0,0)
        surf_size(s,&nv,&ne,&nf);
        printf("  x verts: %d\n",nv); fflush(stdout);
        COMPVERTS( 0,-1, 0,0,1,0,1)
        surf_size(s,&nv,&ne,&nf);
        printf("  x+y verts: %d\n",nv); fflush(stdout);
        COMPVERTS( 0, 0,-1,0,0,1,2)
        surf_size(s,&nv,&ne,&nf);
        printf("  x+y+z verts: %d\n",nv); fflush(stdout);

        // create edges
        //  for each vert 
        //    for each of the four neighbor zone verts (on "+" side)
        //      create edge if it doesn't exist
        //      add to vert's edge list
        dxi=WVERTI(1,0,0,0)-WVERTI(0,0,0,0);
        dyi=WVERTI(0,1,0,0)-WVERTI(0,0,0,0);
        dzi=WVERTI(0,0,1,0)-WVERTI(0,0,0,0);
        #define EDGEGEN(UI,VI,WI,USIZE,VSIZE,WSIZE,D0,D1,D2,DU,DV,DW) {       \
            for (UI= -1;UI<USIZE;UI++) {                                      \
                for (VI=0;VI<VSIZE;VI++) {                                    \
                    for (WI=0;WI<WSIZE;WI++) {                                \
                        i=WVERTI(xi,yi,zi,D0);                                \
                        if (v=vhash_find(i)) {                                \
                            getsample(xi,yi,zi,&x0,&y0,&z0,&d0);              \
                            rev=(d0>=isolevel?0:1);                           \
                            i-=D0;                                            \
                            addedgesub(s,rev,v,D1+i   ,D0+i+DV,D1+i+DU   );   \
                            addedgesub(s,rev,v,D2+i   ,D0+i+DW,D2+i+DU   );   \
                            addedgesub(s,rev,v,D1+i-DV,D0+i-DV,D1+i+DU-DV);   \
                            addedgesub(s,rev,v,D2+i-DW,D0+i-DW,D2+i+DU-DW);   \
                        }                                                     \
                    }                                                         \
                }                                                             \
            }                                                                 \
        }
        EDGEGEN(xi,yi,zi,xsize,ysize,zsize,0,1,2,dxi,dyi,dzi)
        surf_size(s,&nv,&ne,&nf);
        printf("  x edges: %d\n",ne); fflush(stdout);
        EDGEGEN(yi,zi,xi,ysize,zsize,xsize,1,2,0,dyi,dzi,dxi)
        surf_size(s,&nv,&ne,&nf);
        printf("  x+y edges: %d\n",ne); fflush(stdout);
        EDGEGEN(zi,xi,yi,zsize,xsize,ysize,2,0,1,dzi,dxi,dyi)
        surf_size(s,&nv,&ne,&nf);
        printf("  x+y+z edges: %d\n",ne); fflush(stdout);

        // de-allocate temp tables
        mem_put((char *)vhashtab,"bof2mes");
        mempool_destroy(vhmp);

        // fill in faces
        {
            int ei,saw_dummy;
            edge e0,e1,*edges;

            surf_size(s,&nv,&ne,&nf);
            edges=(edge *)mem_get(ne*sizeof(edge),"main");
            ei=0; BGNEDGES edges[ei++]=(e<e->e?e:(edge)0); ENDEDGES

            // add face for faceless edges that have no v_dummy in face walk
            for (ei=0;ei<ne;ei++) {
                if (ex=edges[ei]) {
                    if (!ex->f) {
                        saw_dummy=0;
                        for (e0=e=ex;e;e=e1) {
                            e1=e->e->ve0; if (e1==e0) e1=(edge)0;
                            if (e->v==v_dummy || e->e->v==v_dummy)
                                { saw_dummy=1; e1=(edge)0; }
                        }
                        if (!saw_dummy) f=face_create(ex);
                    }
                }
            }

            // eliminate dummy edges, sharpen verts/edges as needed
            for (ei=0;ei<ne;ei++) {
                if (ex=edges[ei]) {
                    if (ex->v==v_dummy || ex->e->v==v_dummy) edge_destroy(ex);
                    else { if (!ex->f || !ex->e->f) edge_sharpen(ex); }
                }
            }
            BGNVERTS
                for (i=0,e0=e=v->e0;e;e=e1) {
                    e1=e->ve1; if (e1==e0) e1=(edge)0;
                    i++;
                }
                if (i==2) vert_sharpen(v);
            ENDVERTS

            mem_put((char *)edges,"main");
        }

        vert_destroy(v_dummy);
    }

    // write the result in the requested format
    printf("write the result in the requested format\n"); fflush(stdout);
    mem_put((char *)brick,"main");
    surf_size(s,&nv,&ne,&nf);
    printf("\nsurf size: nv=%d ne=%d nf=%d [mem=%d]\n",nv,ne,nf,mem_use());
    for (i=0;surffile[i];i++) ;
    i--;
    switch (surffile[i]) {
    case 'j':
        surf_macwrite(s,surffile);
        break;
    case 's':
        surf_write(s,surffile);
        break;
    case 'b':
        surf_mesbwrite(s,surffile);
        break;
    default:
        fprintf(stderr,"*** unknown format--not saving output\n");
        exit(1);
        break;
    }

};


void getsample(int xi,int yi,int zi,
    double *rx,double *ry,double *rz,double *rd)
{
    double x,y,z,d;

    if (shrink>=0) {
        x=qq0+qq1*(double)xi; y=qq0+qq1*(double)yi; z=qq0+qq1*(double)zi;
    }else{
fprintf(stderr,"*** problem for now with no shrink factor\n"); exit(1);
        x=qq*((double)xi-xc); y=qq*((double)yi-yc); z=qq*((double)zi-zc);
    }
    if (xi<0 || xi>=xsize || yi<0 || yi>=ysize || zi<0 || zi>=zsize) {
        fprintf(stderr,"getsample out of bounds\n"); exit(1);
    }else{
        d=(double)brick[(zi*ysize+yi)*xsize+xi];
    }
    *rx=x; *ry=y; *rz=z; *rd=d;
}

vert vhash_find(int i)
{
    int h;
    vhash vh;

    h=i&(hmax-1);
    for (vh=vhashtab[h];vh && vh->i!=i;vh=vh->vh1) ;
    return (vh?vh->v:(vert)0);
}

void addedgesub(surf s,int rev,vert v,int i1,int i2,int i3)
{
    int h,i;
    vert v1,va,vb,vc;
    edge e,e0,e1,ex;
    double dx,dy,dz;

    // lookup verts
    va=vhash_find(i1); vb=vhash_find(i2); vc=vhash_find(i3);

    // figure out v1
    if (rev) v1=(vc?vc:(vb?vb:va));
    else v1=(va?va:(vb?vb:vc));

    if (!v1) {
        // put in dummy edge as placeholder on boundary
        e=edge_create(v,v_dummy);
    }else{

        // see if edge exists (look at v1's edge list)
        ex=(edge)0;
        for (e=e0=v1->e0;e;e=e1) {
            e1=e->ve1; if (e1==e0) e1=(edge)0;
            if (e->e->v==v) ex=e->e;
        }
        if (v1->e0 && !ex)
            { fprintf(stderr,"addedgesub: v1-v edge not found\n"); exit(1); }
        e=ex;

        // if it doesn't exist, create it
        if (!e) {
            dx=v->x-v1->x; if (dx<0.0) dx= -dx;
            dy=v->y-v1->y; if (dy<0.0) dy= -dy;
            dz=v->z-v1->z; if (dz<0.0) dz= -dz;
            if (dx<1e-6) i=(int)floor(v->x+0.5);
            else{
                if (dy<1e-6) i=(int)floor(v->y+0.5);
                else i=(int)floor(v->z+0.5);
            }
            h=0; while ((i&1)==0 && h<5) { i>>=1; h++; }
            edge_enstate(h<<1);
            e=edge_create(v,v1);
        }
    }

    // add the edge to the vertex
    if (rev) vert_insertedge(e,(edge)0,v->e0);
    else vert_addedge(v,e);
}

void bofshrink(int shrink,
    float **rbof,int *rxsize,int *rysize,int *rzsize)
{
    for (;shrink>0;shrink--) bofshrinksub(rbof,rxsize,rysize,rzsize);
}

void bofshrinksub(float **rbof,int *rxsize,int *rysize,int *rzsize)
{
    int sx,sy,sz,sxsize,sysize,szsize;
    int dx,dy,dz,dxsize,dysize,dzsize;
    int di,si,kx,ky,kz;
    int v0,v1,v2,v3,v4,v5,v6,v7;
    float *sbof,*dbof;

    sbof= *rbof; sxsize= *rxsize; sysize= *rysize; szsize= *rzsize;
    dxsize=sxsize>>1; dysize=sysize>>1; dzsize=szsize>>1;
    printf("shrinking to %dx%dx%d\n",dxsize,dysize,dzsize); fflush(stdout);
    dbof=(float *)mem_get(dxsize*dysize*dzsize*sizeof(float),"bofshrinksub");
    kx=1; ky=kx*sxsize; kz=ky*sysize;
    for (dz=0,di=0;dz<dzsize;dz++) {
        for (dy=0;dy<dysize;dy++) {
            for (dx=0;dx<dxsize;dx++,di++) {
                sx=dx<<1; sy=dy<<1; sz=dz<<1;
                si=(sz*sysize+sy)*sxsize+sx;
                v0=sbof[si         ];
                v1=sbof[si      +kx];
                v2=sbof[si   +ky   ];
                v3=sbof[si   +ky+kx];
                v4=sbof[si+kz      ];
                v5=sbof[si+kz   +kx];
                v6=sbof[si+kz+ky   ];
                v7=sbof[si+kz+ky+kx];
                if (v1>v0) v0=v1;
                if (v2>v0) v0=v2;
                if (v3>v0) v0=v3;
                if (v4>v0) v0=v4;
                if (v5>v0) v0=v5;
                if (v6>v0) v0=v6;
                if (v7>v0) v0=v7;
                //dbof[di]=(v0+v1+v2+v3+v4+v5+v6+v7)/8.0;
                dbof[di]=v0;
            }
        }
    }
    if (!dont_put_brick) mem_put((char *)sbof,"bofshrinksub");
    dont_put_brick=0;
    *rbof=dbof; *rxsize=dxsize; *rysize=dysize; *rzsize=dzsize;
}


